
// Graphics supporting Class for Turbo Pascal programmers
// falcosoft - 2005. The unit requires DelphiX (Directx for Delphi) to compile. 

unit Graph;

interface

uses
  windows, Graphics, classes, dib, dxdraws, directx, math;

procedure GetDirectXVersion(major, minor: pbyte);
function GetFreeSysMem: longword; //return value in KB
function GetFreeVidMem: longword; //return value in KB
function GetProcSpeed: longword; //return value in Mhz
function TpRect(aLeft, aTop, aRight, aBottom: integer): trect;
//All rect structures in this unit are windows style left,bottom exclusive. -> rect(0,0,320,240) is 320x240 rectangle.
// If you want TP style inclusive rects use TpRect() function. -> tprect(0,0,319,239) is 320x240 rectangle.
function Convto32(color: integer): integer; register; // convert tcolor,colorref value for 24 or 32 bit surface.
function Convto16(color: integer): word; register; // same for 16 bit 5,6,5 surface
function Convto15(color: integer): word; register; // same for 16 bit 5,5,5 surface

const
  //SetLineStyle
  SolidLn = integer(0);
  DottedLn = integer(2);
  CenterLn = integer(3);
  DashedLn = integer(1);
  DashDotDotLn = integer(4);
  NullLn = integer(5);

  NormWidth = integer(1);
  DoubleWidth = integer(2);
  TripleWidth = integer(3);
  QuadWidth = integer(4);
  ThickWidth = TripleWidth;

  //SetFillStyle
  EmptyFill = integer(1);
  SolidFill = integer(0);
  LineFill = integer(2);
  ColFill = integer(3);
  HatchFill = integer(6);
  SlashFill = integer(5);
  BkSlashFill = integer(4);
  XHatchFill = integer(7);
  UserFill = integer(8);

  tpstyle = alternate;
  allpoly = winding;

  //PutImage
  CopyPut = SRCCOPY;
  XorPut = SRCINVERT;
  OrPut = SRCPAINT;
  AndPut = SRCAND;
  NotPut = NOTSRCCOPY;
  NormalPut = CopyPut;

  //SetWriteMode
  CopyMode = integer(4);
  XorMode = integer(14);
  OrMode = integer(10);
  AndMode = integer(12);
  NotMode = integer(5);

  //Bar3D
  TopOn = true;
  TopOff = false;

  //SetTextJustify
  LeftText = integer(0);
  CenterText = integer(1);
  RightText = integer(2);
  TopText = integer(0);
  BottomText = integer(1);
  BaselineText = integer(2);

  //SetTextStyle
  CourierNewFont = integer(0);
  SystemFont = integer(1);
  ArialFont = integer(2);
  TimesNewRomanFont = integer(3);
  SansSerifFont = integer(4);
  LucidaConsoleFont = integer(5);
  SymbolFont = integer(6);
  DefaultFont = CourierNewFont;
  ItalicFont = integer($1000);
  UnderlineFont = integer($0100);
  BoldFont = integer($0010);

  HorizDir = integer(0);
  VertDir = integer(90);

  //FloodFill
  BorderFlood = integer(0);
  SurfaceFlood = integer(1);

  //GetMaxColor
  MaxColorNum = integer(0);
  BitsPerPixel = integer(1);

  Rad = Pi / 180.0;
  MaxFonts = 16;
  MinCharSize = 8;

type

  DWORDLONG = Int64;

  TMemoryStatusEx = packed record
    dwLength: DWORD;
    dwMemoryLoad: DWORD;
    ullTotalPhys: DWORDLONG;
    ullAvailPhys: DWORDLONG;
    ullTotalPageFile: DWORDLONG;
    ullAvailPageFile: DWORDLONG;
    ullTotalVirtual: DWORDLONG;
    ullAvailVirtual: DWORDLONG;
    ullAvailExtendedVirtual: DWORDLONG;
  end;

  PRGB = ^TRGB;
  TRGB = packed record
    R, G, B: Byte;
  end;

  PointType = tpoint;
  PPoints = ^TPoints;
  TPoints = array[0..0] of TPoint;

  ArcCoordsType = record
    x, y, xstart, ystart, xend, yend: integer;
  end;
  LineSettingsType = record
    linestyle, pattern, thickness: integer;
  end;

  FillPatternType = array[1..8] of byte;

  FillSettingsType = record
    pattern: integer;
    color: longint;
  end;
  TextSettingsType = record
    font, direction, charsize, horiz, vert: integer;
  end;

  TFontString = string[31];

  ViewPortType = record
    x1, y1, x2, y2: integer;
    clip: boolean;
  end;

  TfalcoMemoryStream = class(Tmemorystream)
  public
    procedure SetPointer(Ptr: Pointer; Size: Longint);
  end;

  TPgraph = class(Tobject)
  private
    fddsd: TDDSurfaceDesc;
    fsurfclipper: Tdirectdrawclipper;
    fLastArcCoords: ArcCoordsType;
    fLineSettings: LineSettingsType;
    fFillSettings: FillSettingsType;
    fTextSettings: TextSettingsType;
    ffillPattern: FillPatternType;
    fViewPort: ViewPortType;
    fgrClip: HRGN;
    fbmp: tbitmap;
    ffloodmode: integer;
    fbkcolor: longint;
    fsurfbkcolor: longint;
    fcanvas: tcanvas;
    fdxdraw: tdxdraw;
    fsurfaces: array of tdirectdrawsurface;
    fInstFont: array[0..MaxFonts - 1] of TFontString;

    procedure InstallDefaultFonts;

  public
    constructor Create(canvas: tcanvas; installfonts: boolean = true); overload;
    constructor Create(dxdraw: tdxdraw; maxsurfacecount: integer = 8; installfonts: boolean = true); overload;
    destructor Destroy; override;

    procedure ClearDevice;
    procedure ClearViewPort;
    function GetMaxX: integer;
    function GetMaxY: integer;
    function GetX: integer;
    function GetY: integer;
    procedure GetViewSettings(var viewport: ViewPortType);
    procedure SetViewPort(x1, y1, x2, y2: integer; clip: boolean);

    function GetBkColor: longint;
    function GetMaxColor(coltype: integer = maxcolornum): longword;
    function SetBkColor(color: longint): longint;
    function GetColor: longint;
    procedure SetColor(color: longint);
    function GetPixel(x, y: integer): longint;
    function PutPixel(x, y: integer; color: longint): longint;
    procedure GetImage(srcrect: trect; bitmap: tgraphic); overload; //for Tdib and Tbitmap
    procedure GetImage(x1, y1, x2, y2: integer; var bitmap); overload;
    function ImageSize(x1, y1, x2, y2: integer): longint;
    procedure PutImage(dstpoint: tpoint; bitmap: tgraphic; bit: longint); overload; //for Tdib and Tbitmap
    procedure PutImage(x1, y1: integer; var bitmap; bit: longint); overload;
    // ! these methods work only if constructor is called with a tdxdraw parameter
    function ColorMatch(Color: Tcolor): longint;
    function PalColor(index: byte): integer;
    procedure SetRGBPalette(nrcolor, r, g, b: byte);
    procedure CreateSurface(surfaceindex: integer; systemmemory: boolean = false);
    procedure FreeSurface(surfaceindex: integer);
    procedure GetSurfImage(surfaceindex: integer; srcrect: trect);
    procedure LoadSurfImage(surfaceindex: integer; srcgraphic: tgraphic); overload;
    procedure LoadSurfImage(surfaceindex: integer; srcfilename: string); overload;
    procedure PutSurfImage(surfaceindex: integer; x, y: integer; transparent: boolean = false; transparentcol: longint = 0); overload;
    procedure PutSurfImage(surfaceindex: integer; destrect, srcrect: trect; transparent: boolean = false; transparentcol: longint = 0); overload;
    // Rot,Wave image methods don't support Viewport clipping.
    procedure PutSurfImageRot(surfaceindex: integer; x, y: integer; centerx, centery: double; angle: integer; transparent: boolean = false; transparentcol: longint = 0; alpha: integer = 255); overload;
    procedure PutSurfImageRot(surfaceindex: integer; destrect, srcrect: trect; centerx, centery: double; angle: integer; transparent: boolean = false; transparentcol: longint = 0; alpha: integer = 255); overload;
    procedure PutSurfImageWave(surfaceindex: integer; x, y: integer; amp, length, phase: integer; transparent: boolean = false; transparentcol: longint = 0; alpha: integer = 255); overload;
    procedure PutSurfImageWave(surfaceindex: integer; destrect, srcrect: trect; amp, length, phase: integer; transparent: boolean = false; transparentcol: longint = 0; alpha: integer = 255); overload;

    procedure ReleaseCanvas;
    procedure SetVisualPage(dummy: integer = 0); //there's no real page parameter

    procedure StartPixelS; // These methods must be called respectively before and after the loop(s) that use GetPixelS,PutPixelS,
    procedure StopPixelS; //don't place other drawing methods in the loops.
    function GetPixelS(x, y: integer): longint; // Faster pixel methods for 8,16,24,32 bit colors.
    procedure PutPixelS(x, y: integer; surfcolor: longint); // Color must be matched by ColorMatch.
    // !

    procedure Arc(x, y: integer; start, stop, radius: integer);
    procedure Circle(x, y: integer; radius: integer);
    procedure DrawPoly(nrpoints: integer; var polypoints); overload;
    procedure DrawPoly(const polypoints: array of tpoint); overload;
    procedure DrawBezier(nrpoints: integer; var polypoints); overload;
    procedure DrawBezier(const polypoints: array of tpoint); overload;
    procedure Ellipse(x, y: integer; start, stop, xradius, yradius: integer);
    procedure GetArcCoords(var arccoords: ArcCoordsType);
    procedure GetLineSettings(var lineinfo: LineSettingsType);
    procedure Line(x1, y1, x2, y2: integer);
    procedure LineRel(dx, dy: integer);
    procedure LineTo(x, y: integer);
    procedure MoveRel(dx, dy: integer);
    procedure MoveTo(x, y: integer);
    procedure Rectangle(x1, y1, x2, y2: integer);
    procedure RotEllipse(x, y, rot: integer; xradius, yradius: integer);
    procedure SetLineStyle(linestyle, pattern, thickness: integer); //! pattern always ignored
    procedure SetWriteMode(writemode: integer);

    procedure Bar(x1, y1, x2, y2: integer);
    procedure Bar3D(x1, y1, x2, y2: integer; depth: integer; top: boolean);
    procedure Chord(x, y: integer; start, stop, xradius, yradius: integer);
    procedure FillBezier(nrpoints: integer; var polypoints); overload;
    procedure FillBezier(const polypoints: array of tpoint); overload;
    procedure FillEllipse(x, y: integer; xradius, yradius: integer);
    procedure FillPoly(nrpoints: integer; var polypoints); overload;
    procedure FillPoly(const polypoints: array of tpoint); overload;
    procedure FillRect(x1, y1, x2, y2: integer);
    procedure FillRotEllipse(x, y, rot: integer; xradius, yradius: integer);
    procedure FloodFill(x, y: integer; color: longint);
    procedure GetFillPattern(var fillpattern: FillPatternType);
    procedure GetFillSettings(var fillinfo: FillSettingsType);
    procedure InvertRect(x1, y1, x2, y2: integer);
    procedure PieSlice(x, y: integer; start, stop, radius: integer);
    procedure RoundRect(x1, y1, x2, y2, cornerx, cornery: integer);
    procedure Sector(x, y: integer; start, stop, xradius, yradius: integer);
    procedure SetFillPattern(fillpattern: FillPatternType; color: longint); overload;
    procedure SetFillPattern(fillpattern: tgraphic); overload; //for 8x8 Tdib and Tbitmap
    procedure SetFillStyle(pattern: integer; color: longint; polyfillmode: integer = tpstyle);
    procedure SetFloodMode(floodmode: integer);

    procedure GetTextSettings(var textinfo: TextSettingsType);
    function InstallUserFont(const fontname: shortstring): integer;
    procedure OutText(const textstring: shortstring);
    procedure OutTextXY(x, y: integer; const textstring: shortstring);
    procedure SetTextJustify(horiz, vert: integer);
    procedure SetTextStyle(font, direction, charsize: integer);
    function TextHeight(const textstring: shortstring): integer;
    function TextWidth(const textstring: shortstring): integer;

  end;

implementation

function GlobalMemoryStatusEx(var lpBuffer: TMemoryStatusEx): BOOL;
type
  TGlobalMemoryStatusEx = function(var lpBuffer: TMemoryStatusEx): BOOL; stdcall;
var
  hkernel: HMODULE;
  GlobMemEx: TGlobalMemoryStatusEx;
begin
  result := false;
  hkernel := LoadLibrary('kernel32.dll');
  if hkernel <> 0 then
  begin
    try
      @GlobMemEx := GetProcAddress(hkernel, 'GlobalMemoryStatusEx');
      if @GlobMemEx <> nil then
      begin
        result := GlobMemEx(lpBuffer);
      end;
    finally
      FreeLibrary(hkernel);
    end;
  end;
end;

procedure GetDirectXVersion(major, minor: pbyte);
var
  key: hkey;
  size: Longword;
  DXVersion: array[0..64] of char;
  code: integer;

begin
  fillchar(DxVersion, 64, 0);
  size := 64;
  regopenkeyex(HKEY_LOCAL_MACHINE, 'Software\Microsoft\DirectX', 0, key_read, key);
  regqueryvalueex(key, 'Version', nil, nil, @dxversion, @size);
  if assigned(major) then
    val(copy(dxversion, 3, 2), major^, code);
  if assigned(minor) then
    val(copy(dxversion, 6, 2), minor^, code);
end;

function GetFreeSysMem: Longword;
var
  ms: tmemorystatus;
  msEx: TMemoryStatusEx;
begin
  FillChar(ms, SizeOf(ms), 0);
  FillChar(msEx, SizeOf(msEx), 0);
  msEx.dwLength := SizeOf(msEx);
  ms.dwLength := sizeof(ms);
  if not GlobalMemoryStatusEx(msEx) then
    globalmemorystatus(ms);

  if msEx.ullTotalPhys > 0 then
    result := msEx.ullAvailPhys div 1024
  else
    result := ms.dwAvailPhys div 1024;

end;

function GetFreeVidMem: Longword;
var
  ddraw: idirectdraw;
  ddraw2: idirectdraw2;
  ddsc: tddscaps;
  totm, freem: dword;
begin
  freem := 0;
  DirectDrawCreate(nil, ddraw, nil);
  ddraw2 := ddraw as idirectdraw2;
  ddsc.dwCaps := DDSCAPS_LOCALVIDMEM;
  DDraw2.GetAvailableVidMem(ddsc, totm, freem);
  result := freem div 1024;
  ddraw2 := nil;
  ddraw := nil;
end;

function GetProcSpeed: longword;
var
  su1, su2, tim1, tim2, freq: int64;

  function tsccall: int64;
  asm
 db $0F,$31               /// rdtsc
  end;

begin
  SetPriorityClass(getcurrentprocess, REALTIME_PRIORITY_CLASS);
  QueryPerformanceFrequency(freq);
  su1 := tsccall;
  queryperformancecounter(tim1);
  repeat
    queryperformancecounter(tim2);
  until (tim1 + (freq * 0.05)) <= tim2;
  su2 := tsccall;
  SetPriorityClass(getcurrentprocess, NORMAL_PRIORITY_CLASS);
  result := round((su2 - su1) * 0.00002);
end;

function TpRect(aLeft, aTop, aRight, aBottom: integer): trect;
begin
  with Result do
  begin
    Left := ALeft;
    Top := ATop;
    Right := ARight + 1;
    Bottom := ABottom + 1;
  end;
end;

function Convto32(color: integer): integer; register;
asm
mov ecx,eax
rol eax,16
and ecx,$0000FF00
and eax,$00FF00FF
or eax,ecx
end;

function Convto16(Color: Integer): Word; register;
asm
 mov ecx,eax
 shl eax,24
 shr eax,27
 shl eax,11
 mov edx,ecx
 shl edx,16
 shr edx,26
 shl edx,5
 or eax,edx
 mov edx,ecx
 shl edx,8
 shr edx,27
 or eax,edx
end;

function Convto15(Color: Integer): Word; register;
asm
 mov ecx,eax
 shl eax,24
 shr eax,27
 shl eax,10
 mov edx,ecx
 shl edx,16
 shr edx,27
 shl edx,5
 or eax,edx
 mov edx,ecx
 shl edx,8
 shr edx,27
 or eax,edx
end;

procedure Tfalcomemorystream.SetPointer;
begin
  inherited;
end;

{ TPgraph }

constructor TPgraph.Create(canvas: tcanvas; installfonts: boolean);
begin
  ffloodmode := 0;
  fbkcolor := 0;
  fdxdraw := nil;
  fsurfaces := nil;
  fcanvas := canvas;
  fbmp := tbitmap.Create;
  fbmp.PixelFormat := pf24bit;
  fbmp.Width := 8;
  fbmp.Height := 8;
  fviewport.x1 := 0;
  fviewport.y1 := 0;
  fviewport.x2 := fcanvas.ClipRect.Right - 1;
  fviewport.y2 := fcanvas.ClipRect.Right - 1;
  fviewport.clip := false;
  fgrclip := 0;
  flinesettings.linestyle := integer(fcanvas.Pen.Style);
  flinesettings.thickness := fcanvas.pen.Width;
  ffillsettings.pattern := integer(fcanvas.brush.Style);
  ffillsettings.color := fcanvas.brush.Color;
  if installfonts then
    InstallDefaultFonts;
end;

constructor TPgraph.Create(dxdraw: tdxdraw; maxsurfacecount: integer; installfonts: boolean);
begin
  ffloodmode := 0;
  fbkcolor := 0;
  fsurfbkcolor := 0;
  fdxdraw := dxdraw;
  fsurfclipper := tdirectdrawclipper.Create(fdxdraw.DDraw);
  SetLength(fsurfaces, maxsurfacecount);
  fcanvas := dxdraw.surface.canvas;
  fbmp := tbitmap.Create;
  fbmp.PixelFormat := pf24bit;
  fbmp.Width := 8;
  fbmp.Height := 8;
  fviewport.x1 := 0;
  fviewport.y1 := 0;
  fviewport.x2 := fdxdraw.SurfaceWidth - 1;
  fviewport.y2 := fdxdraw.SurfaceHeight - 1;
  fviewport.clip := false;
  fgrclip := 0;
  flinesettings.linestyle := integer(fcanvas.Pen.Style);
  flinesettings.thickness := fcanvas.pen.Width;
  ffillsettings.pattern := integer(fcanvas.brush.Style);
  ffillsettings.color := fcanvas.brush.Color;
  if installfonts then
    InstallDefaultFonts;
end;

destructor TPgraph.Destroy;
begin
  fbmp.Free;
  fbmp := nil;
  fsurfaces := nil;
  fsurfclipper.Free;
  if fgrclip <> 0 then
    Deleteobject(fgrclip);
end;

procedure TPgraph.InstallDefaultFonts;
const
  NrDefFonts = 7;
  DefaultFont: array[0..NrDefFonts - 1] of TFontString =
  ('Courier New', 'System', 'Arial', 'Times New Roman',
    'MS Sans Serif', 'Lucida Console', 'Symbol');
var
  i: longint;
begin
  for i := 0 to MaxFonts - 1 do
    fInstFont[i] := '';
  for i := 0 to NrDefFonts - 1 do
  begin
    InstallUserFont(DefaultFont[i]);
  end;
end;

procedure TPgraph.Bar(x1, y1, x2, y2: integer);
begin
  windows.FillRect(fcanvas.handle, rect(x1, y1, x2 + 1, y2 + 1), fcanvas.Brush.Handle);
end;

procedure TPgraph.Bar3D(x1, y1, x2, y2: integer; depth: integer; top: boolean);
var
  pt: array[1..4] of TPOINT;
begin
  windows.Rectangle(fcanvas.Handle, x1, y1, x2 + 1, y2 + 1);

  if top then
  begin
    pt[1].x := x1;
    pt[1].y := y1;
    pt[2].x := x1 + depth;
    pt[2].y := y1 - depth;
    pt[3].x := x2 + depth;
    pt[3].y := y1 - depth;
    pt[4].x := x2;
    pt[4].y := y1;
    windows.Polyline(fcanvas.handle, pt, 4);
  end;
  if (depth <> 0) then
  begin
    pt[1].x := x2 + depth;
    pt[1].y := y1 - depth;
    pt[2].x := x2 + depth;
    pt[2].y := y2 - depth;
    pt[3].x := x2;
    pt[3].y := y2;
    windows.Polyline(fcanvas.handle, pt, 3);
  end;
end;

procedure TPgraph.Chord(x, y: integer; start, stop, xradius,
  yradius: integer);

var
  nXRadial1, nYRadial1, nXRadial2, nYRadial2: integer;
begin

  nXRadial1 := Round(xradius * Cos(start * Rad));
  nXRadial2 := Round(xradius * Cos(stop * Rad));
  nYRadial1 := Round(yradius * Sin(start * Rad));
  nYRadial2 := Round(yradius * Sin(stop * Rad));
  windows.Chord(fcanvas.handle, x - xradius, y - yradius, x + xradius + 1, y + yradius + 1, x + nXRadial1,
    y - nYRadial1, x + nXRadial2, y - nYRadial2);
end;

procedure TPgraph.FillEllipse(x, y: integer; xradius, yradius: integer);
begin
  windows.Ellipse(fcanvas.handle, x - xradius, y - yradius, x + xradius + 1, y + yradius + 1);
end;

procedure TPgraph.FillPoly(nrpoints: integer; var polypoints);
begin
  Windows.Polygon(fcanvas.Handle, PPoints(@polypoints)^, nrpoints);
end;

procedure TPgraph.FillPoly(const polypoints: array of tpoint);
begin
  Windows.Polygon(fcanvas.Handle, PPoints(@polypoints)^, High(polypoints) + 1);
end;

procedure TPgraph.FillRect(x1, y1, x2, y2: integer);
begin
  windows.Rectangle(fcanvas.Handle, x1, y1, x2 + 1, y2 + 1);
end;

procedure TPgraph.FloodFill(x, y: integer; color: longint);
begin
  windows.ExtFloodFill(fcanvas.handle, x, y, color, FFloodMode);
end;

procedure TPgraph.PieSlice(x, y: integer; start, stop, radius: integer);
begin
  Sector(x, y, start, stop, radius, radius);
end;

procedure TPgraph.Sector(x, y: integer; start, stop, xradius,
  yradius: integer);
var
  nXRadial1, nYRadial1, nXRadial2, nYRadial2: integer;
begin

  nXRadial1 := Round(xradius * Cos(start * Rad));
  nXRadial2 := Round(xradius * Cos(stop * Rad));
  nYRadial1 := Round(yradius * Sin(start * Rad));
  nYRadial2 := Round(yradius * Sin(stop * Rad));
  windows.pie(fcanvas.Handle, x - xradius, y - yradius, x + xradius + 1, y + yradius + 1, x + nXRadial1,
    y - nYRadial1, x + nXRadial2, y - nYRadial2);
end;

procedure TPgraph.SetFillPattern(fillpattern: FillPatternType; color: longint);
var
  i, j: integer;
  b: byte;
  col0, col1: longword;
begin
  col1 := colortorgb(color);
  col0 := colortorgb(fBkColor);
  for i := 0 to 7 do
  begin
    b := fillpattern[i + 1];
    for j := 7 downto 0 do
    begin
      if (b and $01 <> 0) then
        SetPixel(fbmp.Canvas.Handle, j, i, col1)
      else
        SetPixel(fbmp.Canvas.Handle, j, i, col0);
      b := b shr 1;
    end;
  end;
  fFillSettings.pattern := userfill;
  fFillSettings.color := color;
  ffillpattern := fillpattern;
  fcanvas.brush.Bitmap := fbmp;
end;

procedure TPgraph.SetFillPattern(fillpattern: tgraphic);
begin
  fbmp.Canvas.Draw(0, 0, fillpattern);
  fFillSettings.pattern := userfill;
  fcanvas.brush.Bitmap := fbmp;
end;

procedure TPgraph.SetFillStyle(pattern: integer; color: longint; polyfillmode: integer);
begin
  fFillSettings.pattern := pattern;
  if pattern = userfill then
    fcanvas.brush.Bitmap := fbmp
  else
  begin
    fcanvas.Brush.color := Color;
    fcanvas.Brush.Style := TBrushStyle(pattern);
    fFillSettings.color := color;
  end;
  setpolyfillmode(fcanvas.handle, polyfillmode);
end;

procedure TPgraph.Arc(x, y: integer; start, stop, radius: integer);
begin
  Ellipse(x, y, start, stop, radius, radius);
end;

procedure TPgraph.Circle(x, y: integer; radius: integer);
begin
  Ellipse(x, y, 0, 360, radius, radius);
end;

procedure TPgraph.DrawPoly(nrpoints: integer; var polypoints);
begin
  windows.Polyline(fcanvas.Handle, PPoints(@polypoints)^, nrpoints);
end;

procedure TPgraph.DrawPoly(const polypoints: array of tpoint);
begin
  Windows.Polyline(fcanvas.Handle, PPoints(@polypoints)^, High(polypoints) + 1);
end;

procedure TPgraph.Ellipse(x, y: integer; start, stop, xradius, yradius: integer);
var
  nXStartArc, nYStartArc, nXEndArc, nYEndArc: integer;

begin
  fLastArcCoords.x := x;
  fLastArcCoords.y := y;
  nXStartArc := Round(xradius * Cos(start * Rad));
  nXEndArc := Round(xradius * Cos(stop * Rad));
  nYStartArc := Round(yradius * Sin(start * Rad));
  nYEndArc := Round(yradius * Sin(stop * Rad));
  windows.Arc(fcanvas.Handle, x - xradius, y - yradius, x + xradius + 1, y + yradius + 1, x + nXStartArc,
    y - nYStartArc, x + nXEndArc, y - nYEndArc);

  with fLastArcCoords do
  begin
    xstart := x + nXStartArc;
    ystart := y - nYStartArc;
    xend := x + nXEndArc;
    yend := y - nYEndArc;
  end;
end;

procedure TPgraph.Line(x1, y1, x2, y2: integer);
var
  oldpos: tpoint;
begin
  windows.MoveToEx(fcanvas.handle, x1, y1, @oldpos);
  windows.LineTo(fcanvas.Handle, x2, y2);
  windows.MoveToEx(fcanvas.handle, oldpos.x, oldpos.y, nil);
end;

procedure TPgraph.LineTo(x, y: integer);
begin
  windows.LineTo(fcanvas.Handle, x, y);
end;

procedure TPgraph.MoveTo(x, y: integer);
begin
  windows.MoveToEx(fcanvas.handle, x, y, nil);
end;

procedure TPgraph.Rectangle(x1, y1, x2, y2: integer);
var
  pt: array[1..5] of tPOINT;
begin

  pt[1].x := x1;
  pt[1].y := y1;
  pt[2].x := x2;
  pt[2].y := y1;
  pt[3].x := x2;
  pt[3].y := y2;
  pt[4].x := x1;
  pt[4].y := y2;
  pt[5].x := x1;
  pt[5].y := y1;
  windows.Polyline(fcanvas.Handle, pt, 5);

end;

procedure TPgraph.SetLineStyle(linestyle, pattern, thickness: integer);
begin
  fcanvas.Pen.Style := tpenstyle(linestyle);
  fcanvas.Pen.Width := thickness;
  fLineSettings.linestyle := linestyle;
  fLineSettings.thickness := thickness;
end;

function TPgraph.GetColor: longint;
begin
  result := fcanvas.Pen.Color;
end;

function TPgraph.GetPixel(x, y: integer): longint;
begin
  result := windows.GetPixel(fcanvas.Handle, x, y);
end;

procedure TPgraph.SetColor(color: longint);
begin
  fcanvas.Pen.Color := color;
  fcanvas.Font.Color := color;
end;

function TPgraph.PutPixel(x, y: integer; color: longint): longint;
begin
  result := windows.SetPixel(fcanvas.handle, x, y, colortorgb(color));
end;

procedure TPgraph.OutTextXY(x, y: integer; const textstring: shortstring);
begin
  fcanvas.TextOut(x, y, textstring);
end;

procedure TPgraph.OutText(const textstring: shortstring);
begin
  fcanvas.TextOut(fcanvas.PenPos.x, fcanvas.PenPos.y, textstring);
end;

procedure TPgraph.GetArcCoords(var arccoords: ArcCoordsType);
begin
  arccoords := fLastArcCoords;
end;

procedure TPgraph.DrawBezier(nrpoints: integer; var polypoints);
begin
  windows.PolyBezier(fcanvas.Handle, PPoints(@polypoints)^, nrpoints);
end;

procedure TPgraph.DrawBezier(const polypoints: array of tpoint);
begin
  windows.PolyBezier(fcanvas.Handle, PPoints(@polypoints)^, High(polypoints) + 1);
end;

procedure TPgraph.RotEllipse(x, y, rot: integer; xradius, yradius: integer);
var
  pt: array[1..7] of tPOINT;
  cosrot, sinrot: double;
  x1, y1, i: longint;
begin
  xradius := Round(1.3333 * xradius);
  cosrot := Cos(rot * Rad);
  sinrot := Sin(rot * Rad);
  pt[1].x := 0;
  pt[1].y := -yradius;
  pt[2].x := xradius;
  pt[2].y := -yradius;
  pt[3].x := xradius;
  pt[3].y := yradius;
  pt[4].x := 0;
  pt[4].y := yradius;
  pt[5].x := -xradius;
  pt[5].y := yradius;
  pt[6].x := -xradius;
  pt[6].y := -yradius;
  pt[7].x := 0;
  pt[7].y := -yradius;
  for i := 1 to 7 do
  begin
    x1 := pt[i].x;
    y1 := pt[i].y;
    pt[i].x := x + Round(x1 * cosrot + y1 * sinrot);
    pt[i].y := y + Round(-x1 * sinrot + y1 * cosrot);
  end;
  windows.PolyBezier(fcanvas.Handle, PPoints(@pt)^, 7);
end;

procedure TPgraph.InvertRect(x1, y1, x2, y2: integer);
begin
  windows.InvertRect(fcanvas.handle, rect(x1, y1, x2 + 1, y2 + 1));
end;

procedure TPgraph.RoundRect(x1, y1, x2, y2, cornerx, cornery: integer);
begin
  windows.RoundRect(fcanvas.handle, x1, y1, x2 + 1, y2 + 1, cornerx, cornery);
end;

procedure TPgraph.SetTextJustify(horiz, vert: integer);
var
  htext, vtext: longword;
begin

  case horiz of
    LeftText: htext := TA_LEFT;
    CenterText: htext := TA_CENTER;
    RightText: htext := TA_RIGHT;
  else
    Exit;
  end;
  case vert of
    TopText: vtext := TA_TOP;
    BottomText: vtext := TA_BOTTOM;
    BaselineText: vtext := TA_BASELINE;
  else
    Exit;
  end;
  fTextSettings.horiz := horiz;
  fTextSettings.vert := vert;
  windows.SetTextAlign(fcanvas.handle, htext or vtext or TA_UPDATECP);
end;

procedure TPgraph.SetTextStyle(font, direction, charsize: integer);
var
  loByte, hiByte: byte;
  nrfont: byte;
  fontname: TFontString;
  lplf: LOGFONT;

begin

  loByte := Lo(font);
  hiByte := Hi(font);
  nrfont := loByte mod $10;
  if (nrfont in [0..MaxFonts - 1]) then
    fontname := fInstFont[nrfont]
  else
    fontname := '';
  if (fontname <> '') then
  begin
    fTextSettings.font := font;
    fTextSettings.direction := direction;
    fTextSettings.charsize := charsize;
    if (charsize <= 5) then
      charsize := charsize * MinCharSize;
    with lplf do
    begin
      lfHeight := charsize;
      lfWidth := 0;
      lfEscapement := 10 * direction;
      lfOrientation := 10 * direction;
      lfWeight := (FW_BOLD - FW_NORMAL) * (loByte div $10) + FW_NORMAL;
      lfItalic := hiByte div $10;
      lfUnderline := hiByte mod $10;
      lfStrikeOut := 0;
      lfCharSet := DEFAULT_CHARSET;
      lfOutPrecision := OUT_DEFAULT_PRECIS;
      lfClipPrecision := CLIP_DEFAULT_PRECIS;
      lfQuality := DEFAULT_QUALITY;
      lfPitchAndFamily := DEFAULT_PITCH or FF_DONTCARE;

      Move(fontname[1], lfFaceName[0], Length(fontname));

    end;

    fcanvas.Font.Handle := CreateFontIndirect(lplf);

  end;
end;

function TPgraph.TextHeight(const textstring: shortstring): integer;
begin
  result := fcanvas.TextHeight(textstring);
end;

function TPgraph.TextWidth(const textstring: shortstring): integer;
begin
  result := fcanvas.TextWidth(textstring);
end;

procedure TPgraph.LineRel(dx, dy: integer);
begin
  windows.LineTo(fcanvas.handle, fcanvas.penpos.x + dx, fcanvas.penpos.y + dy);
end;

procedure TPgraph.MoveRel(dx, dy: integer);
begin
  windows.MoveToEx(fcanvas.handle, fcanvas.penpos.x + dx, fcanvas.penpos.y + dy, nil);
end;

function TPgraph.InstallUserFont(const fontname: shortstring): integer;
var
  i: integer;

begin
  Result := -1;

  for i := 0 to MaxFonts - 1 do
    if (fInstFont[i] = '') then
    begin
      fInstFont[i] := fontname + #0;
      Result := i;
      Break;
    end;

end;

procedure TPgraph.GetTextSettings(var textinfo: TextSettingsType);
begin
  textinfo := fTextSettings;
end;

procedure TPgraph.GetFillSettings(var fillinfo: FillSettingsType);
begin
  fillinfo := fFillSettings;
end;

procedure TPgraph.GetFillPattern(var fillpattern: FillPatternType);
begin
  fillpattern := ffillpattern;
end;

procedure TPgraph.GetLineSettings(var lineinfo: LineSettingsType);
begin
  lineinfo := fLineSettings;
end;

procedure TPgraph.GetImage(srcrect: trect; bitmap: tgraphic);
var
  bmp: tbitmap;
  dib: tdib;
begin
  if bitmap is tbitmap then
  begin
    bmp := bitmap as tbitmap;
    bmp.Width := srcrect.Right - srcrect.Left;
    bmp.Height := srcrect.bottom - srcrect.Top;
    bmp.Canvas.CopyRect(rect(0, 0, bmp.Width, bmp.Height), fcanvas, srcrect);
  end;
  if bitmap is tdib then
  begin
    dib := bitmap as tdib;
    dib.Width := srcrect.Right - srcrect.Left;
    dib.Height := srcrect.bottom - srcrect.Top;
    dib.Canvas.CopyRect(rect(0, 0, dib.Width, dib.Height), fcanvas, srcrect);
  end;
end;

procedure TPgraph.GetImage(x1, y1, x2, y2: integer; var bitmap);
var
  bmp: tbitmap;
  ms: tfalcomemorystream;
begin
  ms := tfalcomemorystream.Create;
  ms.SetPointer(@bitmap, imagesize(x1, y1, x2, y2));
  inc(x2);
  inc(y2);
  bmp := tbitmap.Create;
  bmp.pixelformat := pf24bit;
  bmp.Width := x2 - x1;
  bmp.Height := y2 - y1;
  bmp.Canvas.CopyRect(rect(0, 0, bmp.Width, bmp.Height), fcanvas, rect(x1, y1, x2, y2));
  bmp.SaveToStream(ms);
  ms.Free;
  bmp.Free;
end;

function TPgraph.ImageSize(x1, y1, x2, y2: integer): longint;
begin
  Inc(x2);
  Inc(y2);
  Result := SizeOf(BITMAPFILEHEADER) + SizeOf(BITMAPINFOHEADER) +
    ((x2 - x1 + 7) shr 3) * (y2 - y1) * 24; //hardcoded 24-bit bitmap format
end;

procedure TPgraph.PutImage(dstpoint: tpoint; bitmap: tgraphic; bit: longint);
begin
  fcanvas.copymode := bit;
  fcanvas.draw(dstpoint.x, dstpoint.y, bitmap);
end;

procedure TPgraph.PutImage(x1, y1: integer; var bitmap; bit: longint);
var
  bm, width, height: integer;
begin
  bm := integer(@bitmap);
  width := pinteger(bm + 18)^;
  height := pinteger(bm + 22)^;
  StretchDIBits(fcanvas.handle, x1, y1, width, height, 0, 0, width, height, pointer(bm + 54), pbitmapinfo(bm + 14)^, DIB_RGB_COLORS, bit);
end;

procedure TPgraph.CreateSurface(surfaceindex: integer; systemmemory: boolean);
begin
  if assigned(fdxdraw) then
  begin
    fsurfaces[surfaceindex] := tdirectdrawsurface.Create(fdxdraw.DDraw);
    fsurfaces[surfaceindex].SystemMemory := systemmemory;
  end;
end;

procedure TPgraph.FreeSurface(surfaceindex: integer);
begin
  if assigned(fdxdraw) then
  begin
    if (fsurfaces[surfaceindex].SystemMemory) and (fsurfaces[surfaceindex].Width <> 0) and (fsurfaces[surfaceindex].height <> 0) then
      fsurfaces[surfaceindex].IDDSurface4.PageUnLock(0);
    fsurfaces[surfaceindex].free;
    fsurfaces[surfaceindex] := nil;
  end;
end;

procedure TPgraph.PutSurfImage(surfaceindex: integer; x, y: integer; transparent: boolean; transparentcol: longint);
begin
  if assigned(fdxdraw) then
  begin
    Inc(x, fviewport.x1);
    Inc(y, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].TransparentColor := transparentcol;
    fdxdraw.Surface.Draw(x, y, fsurfaces[surfaceindex], transparent);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

procedure TPgraph.PutSurfImage(surfaceindex: integer; destrect, srcrect: trect; transparent: boolean; transparentcol: longint);
begin
  if assigned(fdxdraw) then
  begin
    Inc(destrect.Left, fviewport.x1);
    Inc(destrect.Top, fviewport.y1);
    Inc(destrect.Right, fviewport.x1);
    Inc(destrect.Bottom, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].TransparentColor := transparentcol;
    fdxdraw.Surface.StretchDraw(destrect, srcrect, fsurfaces[surfaceindex], transparent);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

procedure TPgraph.GetSurfImage(surfaceindex: integer; srcrect: trect);
begin
  if assigned(fdxdraw) then
  begin
    Inc(srcrect.Left, fviewport.x1);
    Inc(srcrect.Top, fviewport.y1);
    Inc(srcrect.Right, fviewport.x1);
    Inc(srcrect.Bottom, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].SetSize((srcrect.Right - srcrect.left), (srcrect.bottom - srcrect.top));
    if fsurfaces[surfaceindex].SystemMemory then
      fsurfaces[surfaceindex].IDDSurface4.PageLock(0);
    fsurfaces[surfaceindex].Draw(0, 0, srcrect, fdxdraw.surface, false);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

procedure TPgraph.LoadSurfImage(surfaceindex: integer; srcgraphic: tgraphic);
begin
  if assigned(fdxdraw) then
    fsurfaces[surfaceindex].LoadFromGraphicRect(srcgraphic, 0, 0, Bounds(0, 0, srcgraphic.Width, srcgraphic.Height));
  if fsurfaces[surfaceindex].SystemMemory then
    fsurfaces[surfaceindex].IDDSurface4.PageLock(0);
end;

procedure TPgraph.LoadSurfImage(surfaceindex: integer; srcfilename: string);
begin
  if assigned(fdxdraw) then
    fsurfaces[surfaceindex].LoadFromFile(srcfilename);
  if fsurfaces[surfaceindex].SystemMemory then
    fsurfaces[surfaceindex].IDDSurface4.PageLock(0);
end;

procedure TPgraph.ClearDevice;
begin
  if assigned(fdxdraw) then
  begin
    fdxdraw.Surface.Clipper := nil;
    fdxdraw.Surface.Canvas.Release;
    fdxdraw.surface.Fill(fsurfbkcolor);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end
  else
  begin
    fcanvas.Brush.Color := fbkcolor;
    fcanvas.Brush.style := bssolid;
    bar(-fviewport.x1, -fviewport.y1, fcanvas.cliprect.right, fcanvas.cliprect.bottom);
    fcanvas.Brush.Color := ffillsettings.color;
    fcanvas.Brush.style := TBrushStyle(ffillsettings.pattern);
  end;
end;

procedure TPgraph.ReleaseCanvas;
begin
  if assigned(fdxdraw) then
    fdxdraw.surface.canvas.release;
end;

procedure TPgraph.SetVisualPage(dummy: integer = 0);
begin
  if assigned(fdxdraw) then
  begin
    fdxdraw.surface.canvas.release;
    fdxdraw.Flip;
  end;
end;

function TPgraph.GetBkColor: longint;
begin
  result := fbkcolor;
end;

function TPgraph.SetBkColor(color: longint): longint;
begin
  result := color;
  if assigned(fdxdraw) then
  begin
    fsurfbkcolor := ColorMatch(color);
    result := fsurfbkcolor;
  end;
  fbkcolor := color;
end;

function TPgraph.ColorMatch(Color: Tcolor): longint;
var
  oldc: Integer;
  oldmode: tpenmode;
begin
  result := color;
  if assigned(fdxdraw) then
  begin
    if (fdxdraw.Surface.SurfaceDesc.ddpfPixelFormat.dwRGBBitCount = 32) or (fdxdraw.Surface.SurfaceDesc.ddpfPixelFormat.dwRGBBitCount = 24) then
      result := convto32(Color)
    else if (fdxdraw.Surface.SurfaceDesc.ddpfPixelFormat.dwRGBBitCount = 16) and (fdxdraw.Surface.SurfaceDesc.ddpfPixelFormat.dwGBitMask = 2016) then
      result := convto16(Color)
    else if (fdxdraw.Surface.SurfaceDesc.ddpfPixelFormat.dwRGBBitCount = 16) then
      result := convto15(Color)
    else
    begin
      color := $02000000 or color;
      oldmode := fcanvas.Pen.Mode;
      fcanvas.Pen.Mode := pmcopy;
      oldc := fdxdraw.surface.canvas.Pixels[0, 0];
      fdxdraw.surface.canvas.pixels[0, 0] := color;
      fdxdraw.surface.canvas.Release;
      Result := fdxdraw.surface.Pixels[0, 0];
      fdxdraw.surface.canvas.Pixels[0, 0] := oldc;
      fcanvas.Pen.Mode := oldmode;
    end;
  end;
end;

procedure TPgraph.SetWriteMode(writemode: integer);
begin
  fcanvas.Pen.mode := tpenmode(writemode);
end;

function TPgraph.GetMaxX: integer;
begin
  if assigned(fdxdraw) then
    result := fdxdraw.SurfaceWidth - 1
  else
    result := fcanvas.ClipRect.Right - 1;
end;

function TPgraph.GetMaxY: integer;
begin
  if assigned(fdxdraw) then
    result := fdxdraw.SurfaceHeight - 1
  else
    result := fcanvas.ClipRect.bottom - 1;
end;

function TPgraph.GetX: integer;
begin
  result := fcanvas.PenPos.x;
end;

function TPgraph.GetY: integer;
begin
  result := fcanvas.PenPos.y;
end;

procedure TPgraph.SetFloodMode(floodmode: integer);
begin
  ffloodmode := floodmode;
end;

function TPgraph.GetMaxColor(coltype: Integer): longword;
var
  ic: hdc;
  bitsp: integer;
begin
  ic := createic('display', nil, nil, nil);
  bitsp := getdevicecaps(ic, bitspixel);
  if coltype = bitsperpixel then
    result := bitsp
  else
  begin
    if bitsp > 24 then
      bitsp := 24;
    result := (1 shl bitsp) - 1;
  end;
  deletedc(ic);
end;

procedure TPgraph.FillBezier(nrpoints: integer; var polypoints);
begin
  beginpath(fcanvas.handle);
  windows.PolyBezier(fcanvas.Handle, PPoints(@polypoints)^, nrpoints);
  endpath(fcanvas.handle);
  StrokeAndFillPath(fcanvas.handle);
end;

procedure TPgraph.FillBezier(const polypoints: array of tpoint);
begin
  beginpath(fcanvas.handle);
  windows.PolyBezier(fcanvas.Handle, PPoints(@polypoints)^, High(polypoints) + 1);
  endpath(fcanvas.handle);
  StrokeAndFillPath(fcanvas.handle);
end;

procedure TPgraph.FillRotEllipse(x, y, rot, xradius, yradius: integer);
begin
  beginpath(fcanvas.handle);
  rotellipse(x, y, rot, xradius, yradius);
  endpath(fcanvas.handle);
  StrokeAndFillPath(fcanvas.handle);
end;

procedure TPgraph.PutSurfImageRot(surfaceindex, x, y: integer; centerx,
  centery: double; angle: integer; transparent: boolean;
  transparentcol: longint; alpha: integer);
begin
  if assigned(fdxdraw) then
  begin
    Inc(x, fviewport.x1);
    Inc(y, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].TransparentColor := transparentcol;
    fdxdraw.Surface.DrawRotateAlpha(x, y, fsurfaces[surfaceindex].Width, fsurfaces[surfaceindex].height, fsurfaces[surfaceindex].ClientRect, fsurfaces[surfaceindex], Centerx, CenterY, transparent, round(-angle / 1.40625), alpha);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

procedure TPgraph.PutSurfImageRot(surfaceindex: integer; destrect,
  srcrect: trect; centerx, centery: double; angle: integer; transparent: boolean; transparentcol: longint; alpha: integer);
begin
  if assigned(fdxdraw) then
  begin
    Inc(destrect.Left, fviewport.x1);
    Inc(destrect.Top, fviewport.y1);
    Inc(destrect.Right, fviewport.x1);
    Inc(destrect.Bottom, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].TransparentColor := transparentcol;
    fdxdraw.Surface.DrawRotateAlpha(destrect.Left, destrect.Top, destrect.right - destrect.left, destrect.bottom - destrect.top, srcrect, fsurfaces[surfaceindex], Centerx, CenterY, transparent, round(-angle / 1.40625), alpha);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

procedure TPgraph.PutSurfImageWave(surfaceindex, x, y, amp, length,
  phase: integer; transparent: boolean; transparentcol: longint;
  alpha: integer);
begin
  if assigned(fdxdraw) then
  begin
    Inc(x, fviewport.x1);
    Inc(y, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].TransparentColor := transparentcol;
    fdxdraw.Surface.DrawWavexAlpha(x, y, fsurfaces[surfaceindex].Width, fsurfaces[surfaceindex].height, fsurfaces[surfaceindex].ClientRect, fsurfaces[surfaceindex], transparent, Amp, Length, round(phase / 1.40625), alpha);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

procedure TPgraph.PutSurfImageWave(surfaceindex: integer; destrect,
  srcrect: trect; amp, length, phase: integer; transparent: boolean; transparentcol: longint; alpha: integer);
begin
  if assigned(fdxdraw) then
  begin
    Inc(destrect.Left, fviewport.x1);
    Inc(destrect.Top, fviewport.y1);
    Inc(destrect.Right, fviewport.x1);
    Inc(destrect.Bottom, fviewport.y1);
    fdxdraw.surface.canvas.release;
    fsurfaces[surfaceindex].TransparentColor := transparentcol;
    fdxdraw.Surface.DrawWavexAlpha(destrect.Left, destrect.Top, destrect.right - destrect.left, destrect.bottom - destrect.top, srcrect, fsurfaces[surfaceindex], transparent, Amp, Length, round(phase / 1.40625), alpha);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end;
end;

function TPgraph.GetPixelS(x, y: integer): longint;
begin
  Inc(x, fviewport.x1);
  Inc(y, fviewport.y1);
  Result := 0;
  if (not fviewport.clip) or ((x <= fviewport.x2) and (y <= fviewport.y2)) then
  begin
    case fddsd.ddpfPixelFormat.dwRGBBitCount of
      8: Result := PByte(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X)^;
      16: Result := PWord(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X * 2)^;
      24: with PRGB(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X * 3)^ do
          Result := R or (G shl 8) or (B shl 16);
      32: Result := PInteger(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X * 4)^;
    end;
  end;
end;

procedure TPgraph.PutPixelS(x, y, surfcolor: Integer);
begin
  Inc(x, fviewport.x1);
  Inc(y, fviewport.y1);
  if (not fviewport.clip) or ((x <= fviewport.x2) and (y <= fviewport.y2)) then
  begin
    case fddsd.ddpfPixelFormat.dwRGBBitCount of

      8: PByte(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X)^ := surfcolor;
      16: PWord(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X * 2)^ := surfcolor;
      24: with PRGB(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X * 3)^ do
        begin
          R := Byte(surfcolor);
          G := Byte(surfcolor shr 8);
          B := Byte(surfcolor shr 16);
        end;
      32: PInteger(Integer(fddsd.lpSurface) + Y * fddsd.lPitch + X * 4)^ := surfcolor;
    end;
  end;
end;

procedure TPgraph.StartPixelS;
begin
  fdxdraw.Surface.Canvas.Release;
  fdxdraw.Surface.Lock(fddsd);
end;

procedure TPgraph.StopPixelS;
begin
  fdxdraw.Surface.UnLock;
  setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
end;

procedure TPgraph.ClearViewPort;
begin
  if assigned(fdxdraw) then
  begin
    fdxdraw.Surface.Canvas.Release;
    fdxdraw.surface.Fillrect(tprect(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2), fsurfbkcolor);
    setviewport(fviewport.x1, fviewport.y1, fviewport.x2, fviewport.y2, fviewport.clip);
  end
  else
  begin
    fcanvas.Brush.Color := fbkcolor;
    fcanvas.Brush.style := bssolid;
    bar(0, 0, fviewport.x2 - fviewport.x1, fviewport.y2 - fviewport.y1);
    fcanvas.Brush.Color := ffillsettings.color;
    fcanvas.Brush.style := TBrushStyle(ffillsettings.pattern);
  end;
end;

procedure TPgraph.GetViewSettings(var viewport: ViewPortType);
begin
  viewport := fviewport;
end;

procedure TPgraph.SetViewPort(x1, y1, x2, y2: integer; clip: boolean);
var
  srect: trect;
begin
  fViewPort.x1 := x1;
  fViewPort.y1 := y1;
  fViewPort.x2 := x2;
  fViewPort.y2 := y2;
  fViewPort.clip := clip;
  setviewportorgex(fcanvas.handle, x1, y1, nil);

  if (fgrClip <> 0) then
  begin
    SelectClipRgn(fcanvas.handle, 0);
    DeleteObject(fgrClip);
    fgrClip := 0;
    if assigned(fdxdraw) then
      fdxdraw.Surface.Clipper := nil;
  end;

  if clip then
  begin
    fgrClip := CreateRectRgn(x1, y1, x2 + 1, y2 + 1);
    SelectClipRgn(fcanvas.handle, fgrClip);
    if assigned(fdxdraw) then
    begin
      srect := rect(x1, y1, x2 + 1, y2 + 1);
      fsurfclipper.SetClipRects(srect);
      fdxdraw.Surface.Clipper := fsurfclipper;

    end;
  end;
end;

function TPgraph.PalColor(index: byte): integer;
begin
  if assigned(fdxdraw) then
    result := $02000000 or rgb(fdxdraw.ColorTable[index].rgbred, fdxdraw.ColorTable[index].rgbGreen, fdxdraw.ColorTable[index].rgbBlue)
  else
    result := paletteindex(index);
end;

procedure TPgraph.SetRGBPalette(nrcolor, r, g, b: byte);
begin
  if assigned(fdxdraw) then
  begin
    fdxdraw.ColorTable[nrcolor].rgbBlue := b;
    fdxdraw.ColorTable[nrcolor].rgbGreen := g;
    fdxdraw.ColorTable[nrcolor].rgbRed := r;
    fdxdraw.ColorTable[nrcolor].rgbReserved := 0;
    fdxdraw.DefColorTable[nrcolor] := fdxdraw.ColorTable[nrcolor];
  end;
end;

end.

